// file      : odb/pgsql/simple-object-statements.cxx
// copyright : Copyright (c) 2005-2015 Code Synthesis Tools CC
// license   : GNU GPL v2; see accompanying LICENSE file

#include <odb/pgsql/simple-object-statements.hxx>

namespace odb
{
  namespace pgsql
  {
    object_statements_base::
    ~object_statements_base ()
    {
    }
  }
}
